import pytest
from pikepdf import Pdf, Stream

import os
import platform
import shutil
from contextlib import suppress
from shutil import copy
import gc

check_refcount = pytest.helpers.check_refcount


def test_split_pdf(resources, outdir):
    q = Pdf.open(resources / "fourpages.pdf")

    for n, page in enumerate(q.pages):
        outpdf = Pdf.new()
        outpdf.pages.append(page)
        outpdf.save(outdir / "page{}.pdf".format(n + 1))

    assert len([f for f in outdir.iterdir() if f.name.startswith('page')]) == 4


def test_empty_pdf(outdir):
    q = Pdf.new()
    with pytest.raises(IndexError):
        q.pages[0]
    q.save(outdir / 'empty.pdf')


def test_delete_last_page(resources, outdir):
    q = Pdf.open(resources / 'graph.pdf')
    del q.pages[0]
    q.save(outdir / 'empty.pdf')


def test_replace_page(resources):
    q = Pdf.open(resources / "fourpages.pdf")
    q2 = Pdf.open(resources / "graph.pdf")

    assert len(q.pages) == 4
    q.pages[1] = q2.pages[0]
    assert len(q.pages) == 4
    assert q.pages[1].Resources.XObject.keys() == \
        q2.pages[0].Resources.XObject.keys()


def test_hard_replace_page(resources, outdir):
    q = Pdf.open(resources / "fourpages.pdf")
    q2 = Pdf.open(resources / "graph.pdf")

    q2_page = q2.pages[0]
    del q2
    q.pages[1] = q2_page

    q2 = Pdf.open(resources / 'sandwich.pdf')
    q2_page = q2.pages[0]
    q.pages[2] = q2_page
    del q2
    del q2_page
    gc.collect()

    q.save(outdir / 'out.pdf')


def test_reverse_pages(resources, outdir):
    q = Pdf.open(resources / "fourpages.pdf")
    qr = Pdf.open(resources / "fourpages.pdf")

    lengths = [int(page.Contents.stream_dict.Length) for page in q.pages]

    qr.pages.reverse()
    qr.save(outdir / "reversed.pdf")

    for n, length in enumerate(lengths):
        assert q.pages[n].Contents.stream_dict.Length == length

    for n, length in enumerate(reversed(lengths)):
        assert qr.pages[n].Contents.stream_dict.Length == length


def test_evil_page_deletion(resources, outdir):
    # str needed for py<3.6
    copy(str(resources / 'sandwich.pdf'), str(outdir / 'sandwich.pdf'))

    src = Pdf.open(outdir / 'sandwich.pdf')
    pdf = Pdf.open(resources / 'graph.pdf')

    assert check_refcount(src, 2)
    pdf.pages.append(src.pages[0])
    assert check_refcount(src, 3)

    del src.pages[0]
    gc.collect()
    assert check_refcount(src, 3)
    (outdir / 'sandwich.pdf').unlink()
    pdf.save(outdir / 'out.pdf')

    del pdf.pages[0]
    pdf.save(outdir / 'out2.pdf')

    del pdf.pages[0]
    pdf.save(outdir / 'out_nopages.pdf')
    del pdf
    gc.collect()
    # Ideally we'd see the check_refcount(src, 2) at this point, but we don't
    # have a way to find out when a PDF can be closed if a page was copied out
    # of it to another PDF


def test_append_all(resources, outdir):
    pdf = Pdf.open(resources / 'sandwich.pdf')
    pdf2 = Pdf.open(resources / 'fourpages.pdf')

    for page in pdf2.pages:
        pdf.pages.append(page)

    assert len(pdf.pages) == 5
    pdf.save(outdir / 'out.pdf')


def test_extend(resources, outdir):
    pdf = Pdf.open(resources / 'sandwich.pdf')
    pdf2 = Pdf.open(resources / 'fourpages.pdf')
    pdf.pages.extend(pdf2.pages)

    assert len(pdf.pages) == 5
    pdf.save(outdir / 'out.pdf')


def test_slice_unequal_replacement(resources, outdir):
    pdf = Pdf.open(resources / 'fourpages.pdf')
    pdf2 = Pdf.open(resources / 'sandwich.pdf')

    assert len(pdf.pages[1:]) != len(pdf2.pages)
    page0_content_len = int(pdf.pages[0].Contents.Length)
    page1_content_len = int(pdf.pages[1].Contents.Length)
    pdf.pages[1:] = pdf2.pages

    assert len(pdf.pages) == 2, "number of pages must be changed"
    pdf.save(outdir / 'out.pdf')
    assert pdf.pages[0].Contents.Length == page0_content_len, \
        "page 0 should be unchanged"
    assert pdf.pages[1].Contents.Length != page1_content_len, \
        "page 1's contents should have changed"


def test_slice_with_step(resources, outdir):
    pdf = Pdf.open(resources / 'fourpages.pdf')
    pdf2 = Pdf.open(resources / 'sandwich.pdf')

    pdf2.pages.extend(pdf2.pages[:])
    assert len(pdf2.pages) == 2
    pdf2_content_len = int(pdf2.pages[0].Contents.Length)

    pdf.pages[0::2] = pdf2.pages
    pdf.save(outdir / 'out.pdf')

    assert all(page.Contents.Length == pdf2_content_len
               for page in pdf.pages[0::2])


@pytest.mark.timeout(1)
def test_self_extend(resources):
    pdf = Pdf.open(resources / 'fourpages.pdf')
    with pytest.raises(ValueError,
            message="source page list modified during iteration"):
        pdf.pages.extend(pdf.pages)


def test_page_contents_add(resources, outdir):
    pdf = Pdf.open(resources / 'graph.pdf')

    stream1 = Stream(pdf, b"q 0.707 -0.707 0.707 0.707 0 0 cm")
    stream2 = Stream(pdf, b"Q")

    pdf.pages[0].page_contents_add(stream1, True)
    pdf.pages[0].page_contents_add(stream2, False)
    pdf.save(outdir / 'out.pdf')



from decimal import Decimal
from math import isclose, isfinite
import sys

from pikepdf import _qpdf as qpdf
from pikepdf import (Pdf, Object, Real, String, Array, Integer, Name, Boolean,
    Null)
from hypothesis import given, strategies as st, example
from hypothesis.strategies import (none, integers, binary, lists, floats,
    characters, recursive, booleans, builds, one_of)
import pytest


encode = qpdf._encode
decode = qpdf._decode


def decode_encode(obj):
    return decode(encode(obj))


def test_bool_involution():
    assert decode_encode(True) == True
    assert decode_encode(False) == False


@given(integers(min_value=-2**31, max_value=(2**31-1)))
def test_integer_involution(n):
    assert decode_encode(n) == n


@given(characters(min_codepoint=0x20, max_codepoint=0x7f))
@example('')
def test_ascii_involution(ascii):
    b = ascii.encode('ascii')
    assert decode_encode(b) == b


@given(characters(min_codepoint=0x0, max_codepoint=0xfef0,
                  blacklist_categories=('Cs',)))
@example('')
def test_unicode_involution(s):
    assert str(encode(s)) == s


@given(binary(min_size=0, max_size=300))
def test_binary_involution(binary):
    assert decode_encode(binary) == binary


@given(integers(-10**12, 10**12), integers(-10**12, 10**12))
def test_integer_comparison(a, b):
    equals = (a == b)
    encoded_equals = (qpdf._encode(a) == qpdf._encode(b))
    assert encoded_equals == equals

    lessthan = (a < b)
    encoded_lessthan = (qpdf._encode(a) < qpdf._encode(b))
    assert lessthan == encoded_lessthan


@given(integers(max_value=9223372036854775807), integers(max_value=9223372036854775807))
def test_crosstype_comparison(a, b):
    equals = (a == b)
    encoded_equals = (a == qpdf._encode(b))
    assert encoded_equals == equals

    lessthan = (a < b)
    encoded_lessthan = (qpdf._encode(a) < b)
    assert lessthan == encoded_lessthan


@given(integers(-10**12, 10**12), integers(0, 12))
def test_decimal_involution(num, radix):
    strnum = str(num)
    if radix > len(strnum):
        strnum = strnum[:radix] + '.' + strnum[radix:]

    d = Decimal(strnum)

    assert Real(d).decode() == d


@given(floats())
def test_decimal_from_float(f):
    d = Decimal(f)
    if isfinite(f) and d.is_finite():
        py_d = Real(d)
        assert isclose(py_d.decode(), d), (d, f.hex())
    else:
        with pytest.raises(ValueError, message=repr(f)):
            Real(f)
        with pytest.raises(ValueError, message=repr(d)):
            Real(d)


@given(lists(integers(-10, 10), min_size=0, max_size=10))
def test_list(array):
    assert decode_encode(array) == array


@given(lists(lists(integers(1,10), min_size=1, max_size=5),min_size=1,max_size=5))
def test_nested_list(array):
    assert decode_encode(array) == array


@given(recursive(none() | booleans(), lambda children: lists(children), max_leaves=20))
def test_nested_list2(array):
    assert decode_encode(array) == array


def test_stack_depth():
    a = [42]
    for n in range(100):
        a = [a]
    rlimit = sys.getrecursionlimit()
    try:
        sys.setrecursionlimit(100)
        with pytest.raises(RecursionError, message="recursion"):
            assert decode_encode(a) == a
        with pytest.raises(RecursionError, message="recursion"):
            encode(a) == encode(a)
        with pytest.raises(RecursionError, message="recursion"):
            repr(a)
    finally:
        sys.setrecursionlimit(rlimit)  # So other tests are not affected


def test_bytes():
    b = b'\x79\x78\x77\x76'
    qs = String(b)
    assert bytes(qs) == b

    s = 'é'
    qs = String(s)
    assert str(qs) == s


def test_len_array():
    assert len(Array([])) == 0
    assert len(Array([Integer(3)])) == 1


class TestHashViolation:

    def check(self, a, b):
        assert a == b, "invalid test case"
        assert hash(a) == hash(b), "hash violation"

    def test_unequal_but_similar(self):
        assert Name('/Foo') != String('/Foo')

    def test_numbers(self):
        self.check(Real('1.0'), Integer(1))
        self.check(Real('42'), Integer(42))

    def test_bool_comparison(self):
        self.check(Real('0.0'), Boolean(0))
        self.check(Boolean(1), Integer(1))

    def test_string(self):
        utf16 = b'\xfe\xff' + 'hello'.encode('utf-16be')
        self.check(String(utf16), String('hello'))


def test_not_constructible():
    with pytest.raises(TypeError, message="constructor"):
        Object()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)


from collections import namedtuple
from enum import Enum
from pkg_resources import get_distribution, DistributionNotFound

import os

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    __version__ = "Not installed"
    pass
try:
    from . import _qpdf
except ImportError:
    raise ImportError("pikepdf's extension library failed to import")

from ._qpdf import (Object, ObjectType, PdfError, Pdf, PasswordError,
        ObjectStreamMode, StreamDataMode)

from ._objects import (Boolean, Integer, Real, Name, String, Array, Dictionary,
        Stream, Operator, Null)

from ._pdfimage import PdfImage, PdfInlineImage, UnsupportedImageTypeError

__libqpdf_version__ = _qpdf.qpdf_version()


def parse_content_stream(page_or_stream, operators=''):
    """Parse a PDF content stream into a sequence of instructions.

    A PDF content stream is list of instructions that describe where to render
    the text and graphics in a PDF. This is the starting point for analyzing
    PDFs.

    If the input is a page and page.Contents is an array, then the content
    stream is automatically treated as one coalesced stream.

    Each instruction contains at least one operator and zero or more operands.

    :param page_or_stream: A page, or a content :class:`pikepdf.Stream` attached
    to another object such as a Form XObject
    :param operators: A space-separated string of operators that the caller is
    inserted in, for example 'q Q cm Do' will return only operators that pertain
    to drawing images. Use 'BI ID EI' for inline images.

    >>> pdf = pikepdf.Pdf.open(input_pdf)
    >>> page = pdf.pages[0]
    >>> for operands, command in parse_content_stream(page):
    >>>     print(command)

    """

    if not isinstance(page_or_stream, Object):
        raise TypeError("stream must a PDF object")

    if page_or_stream.type_code != ObjectType.stream \
            and page_or_stream.get('/Type') != '/Page':
        raise TypeError("parse_content_stream called on page or stream object")

    try:
        if page_or_stream.get('/Type') == '/Page':
            page = page_or_stream
            instructions = page._parse_page_contents_grouped(operators)
        else:
            stream = page_or_stream
            instructions = Object._parse_stream_grouped(stream, operators)
    except PdfError as e:
        if 'parseContentStream called on non-stream' in str(e):  # qpdf 6.x
            raise TypeError("parse_content_stream called on non-stream Object")
        elif 'ignoring non-stream while parsing' in str(e):  # qpdf 7.0
            raise TypeError("parse_content_stream called on non-stream Object")
        raise e from e

    return instructions


class Page:
    def __init__(self, obj):
        self.obj = obj

    def __getattr__(self, item):
        return getattr(self.obj, item)

    def __setattr__(self, item, value):
        if item == 'obj':
            object.__setattr__(self, item, value)
        elif hasattr(self.obj, item):
            setattr(self.obj, item, value)
        else:
            raise AttributeError(item)

    def __repr__(self):
        return repr(self.obj).replace(
            'pikepdf.Dictionary', 'pikepdf.Page', 1)

    @property
    def mediabox(self):
        return self.obj.MediaBox

    def has_text(self):
        """Check if this page print text

        Search the content stream for any of the four text showing operators.
        We ignore text positioning operators because some editors might
        generate maintain these even if text is deleted etc.

        This cannot detect raster text (text in a bitmap), text rendered as
        curves. It also cannot determine if the text is visible to the user.

        :return: True if there is text
        """
        text_showing_operators = """TJ " ' Tj"""
        text_showing_insts = parse_content_stream(
            self.obj, text_showing_operators)
        if len(text_showing_insts) > 0:
            return True
        return False


def open(*args, **kwargs):
    "Alias for :func:`pikepdf.Pdf.open`."
    return Pdf.open(*args, **kwargs)


class PdfMatrix:
    """Support class for PDF content stream matrices

    PDF content stream matrices are 3x3 matrices summarized by a shorthand
    (a, b, c, d, e, f) which correspond to the first two column vectors. The
    final column vector is always (0, 0, 1) since this is using homogenous
    coordinates.

    PDF uses row vectors.  That is vr @ A' gives the effect of transforming
    a row vector vr=(x, y, 1) by the matrix A'.  It's more common to use
    A @ vc where vc is a column vector = (x, y, 1)'.

    Addition is not implemented. If needed it would be necessary to normalize
    with division by self.values[2][2].

    """

    def __init__(self, other):
        if isinstance(other, PdfMatrix):
            self.values = other.values
        elif len(other) == 6:
            a, b, c, d, e, f = map(float, other)
            self.values = ((a, b, 0),
                           (c, d, 0),
                           (e, f, 1))
        elif len(other) == 3 and len(other[0]) == 3:
            self.values = (tuple(other[0]),
                           tuple(other[1]),
                           tuple(other[2]))
        else:
            raise ValueError('arguments')

    @staticmethod
    def identity():
        return PdfMatrix((1, 0, 0, 1, 0, 0))

    def __matmul__(self, other):
        a = self.values
        b = other.values
        return PdfMatrix(
                [[sum([float(i) * float(j)
                       for i, j in zip(row, col)]
                     ) for col in zip(*b)]
                  for row in a]
        )

    @property
    def shorthand(self):
        return (self.a, self.b, self.c, self.d, self.e, self.f)

    @property
    def a(self):
        return self.values[0][0]

    @a.setter
    def a(self, value):
        self.values[0][0] = value

    @property
    def b(self):
        return self.values[0][1]

    @b.setter
    def b(self, value):
        self.values[0][1] = value

    @property
    def c(self):
        return self.values[1][0]

    @c.setter
    def c(self, value):
        self.values[1][0] = value

    @property
    def d(self):
        return self.values[1][1]

    @d.setter
    def d(self, value):
        self.values[1][1] = value

    @property
    def e(self):
        return self.values[2][0]

    @e.setter
    def e(self, value):
        self.values[2][0] = value

    @property
    def f(self):
        return self.values[2][1]

    @f.setter
    def f(self, value):
        self.values[2][1] = value

    def encode(self):
        return '{:.6f} {:.6f} {:.6f} {:.6f} {:.6f} {:.6f}'.format(
            self.a, self.b, self.c, self.d, self.e, self.f
        ).encode()



import pytest
from pikepdf import Pdf, Dictionary


@pytest.fixture
def vera(resources):
    pdf = Pdf.open(resources / 'veraPDF test suite 6-2-10-t02-pass-a.pdf')
    return pdf


@pytest.fixture
def graph(resources):
    return Pdf.open(resources / 'graph.pdf')


def test_no_info(vera, outdir):
    assert vera.trailer.get('/Info') is None, 'need a test file with no /Info'

    assert len(vera.metadata) == 0
    creator = 'pikepdf test suite'
    vera.metadata['/Creator'] = creator
    assert vera.metadata.is_indirect, "/Info must be an indirect object"
    vera.save(outdir / 'out.pdf')

    new = Pdf.open(outdir / 'out.pdf')
    assert new.metadata['/Creator'] == creator


def test_update_info(graph, outdir):
    new_title = '我敢打赌，你只是想看看这意味着什么'
    graph.metadata['/Title'] = new_title
    graph.save(outdir / 'out.pdf')

    new = Pdf.open(outdir / 'out.pdf')
    assert new.metadata['/Title'] == new_title
    assert graph.metadata['/Author'] == new.metadata['/Author']

    with pytest.raises(ValueError):
        new.metadata = Dictionary({'/Keywords': 'bob'})

    new.metadata = graph.make_indirect(Dictionary({'/Keywords': 'bob'}))
    assert new.metadata.is_indirect, "/Info must be an indirect object"


def test_copy_info(vera, graph, outdir):
    vera.metadata = vera.copy_foreign(graph.metadata)
    assert vera.metadata.is_indirect, "/Info must be an indirect object"
    vera.save(outdir / 'out.pdf')



import gc
import pytest
from pikepdf import Pdf

check_refcount = pytest.helpers.check_refcount

# Try to do some things without blowing up

def test_access_image(resources):
    pdf = Pdf.open(resources / 'congress.pdf')
    assert check_refcount(pdf, 2)  # refcount is always +1
    im0 = pdf.pages[0].Resources.XObject['/Im0']
    assert check_refcount(pdf, 3), "didn't acquire a reference to owner"

    del pdf
    gc.collect()
    im0.read_raw_bytes()


def test_access_page(resources):
    pdf = Pdf.open(resources / 'graph.pdf')
    page0 = pdf.pages[0]
    also_page0 = pdf.pages.p(1)
    assert check_refcount(pdf, 4), "didn't acquire a reference to owner"
    del pdf
    gc.collect()
    page0.Contents.read_raw_bytes()
    del page0
    gc.collect()
    also_page0.Contents.read_raw_bytes()


def test_remove_pdf_and_all_pages(resources):
    pdf = Pdf.open(resources / 'graph.pdf')
    page0 = pdf.pages[0]
    contents = page0.Contents
    assert check_refcount(pdf, 4), "stream didn't acquire reference to owner"
    del pdf
    del page0
    gc.collect()
    contents.read_raw_bytes()
    del contents
    gc.collect()


def test_access_pdf_metadata(resources):
    pdf = Pdf.open(resources / 'graph.pdf')
    meta = pdf.Root.Metadata
    del pdf
    gc.collect()
    meta.read_raw_bytes()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)

from . import qpdf
import collections.abc

# This is a repository of half-baked ideas, mainly an effort to automatically
# unbox QPDFObjectHandle to equivalent Python types
# I'm unconvinced that's a good idea in the first place, convenient as it is,
# as it involves roundtripping a lot of data and boxed types can usually have
# a nice repr() or str() or whatever. Also here is a half-baked syntax
# checker.

class PdfObject:
    @classmethod
    def specialize(cls, obj):
        if obj.type_code == qpdf.ObjectType.ot_integer:
            return obj.as_int()  # Integer is transparent

        elif obj.type_code == qpdf.ObjectType.ot_boolean:
            return obj.as_bool()

        elif obj.type_code == qpdf.ObjectType.ot_null:
            return None

        elif obj.type_code == qpdf.ObjectType.ot_dictionary:
            if obj.get("/Type") == qpdf.Object.Name("/Page"):
                return PdfPage(obj)
            return PdfDict(obj)

        return PdfObject(obj)


    def __init__(self, obj):
        self.obj = obj

    def __repr__(self):
        return self.__class__.__name__ + "(" + repr(self.obj) + ")"


class PdfName(PdfObject):
    def __init__(self, name):
        if isinstance(name, str):
            if not name.startswith('/'):
                raise ValueError("Name must start with /")
            self.obj = qpdf.Object.Name(name)
            return

        elif isinstance(name, qpdf.Object):
            self.obj = name
            return

        elif isinstance(name, PdfName):
            self.obj = name.obj
            return

        raise NotImplemented()

    def __str__(self):
        return ""


class PdfDict(PdfObject, collections.abc.MutableMapping):
    def __init__(self, *args, **kwargs):
        if len(args) == 0:
            d = {}
            # Translate PdfDict(Type='/Page', Parent=None...) to
            # a valid name dictionary
            for key, val in kwargs:
                if key[0].isupper():
                    d['/' + key] = val
                else:
                    raise ValueError("Invalid PDF dict /Name")

            self.obj = qpdf.Object.Dictionary(d)
            return

        elif isinstance(args[0], qpdf.Object):
            self.obj = args[0]
            return

        elif isinstance(args[0], collections.abc.Mapping):
            self.obj = qpdf.Object.Dictionary(args[0])
            return

        raise NotImplemented()

    def __contains__(self, key):
        return self.obj.__contains__(key)

    def __len__(self):
        return self.obj.__len__()

    def __iter__(self):
        return iter(self.obj.as_dict())

    def __getitem__(self, key):
        return PdfObject.specialize(self.obj.__getitem__(key))

    def __setitem__(self, key, val):
        return self.obj.__setitem__(key, val)

    def __delitem__(self, key):
        return self.obj.__delitem__(key)

    def keys(self):
        return self.obj.keys()

    def items(self):
        return self.obj.as_dict().items()

    def values(self):
        return self.obj.as_dict().values()


class RestrictedDict(PdfDict):
    REQUIRED_KEYS = frozenset()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def validate(self):
        self_keys = set(self.keys())
        missing_keys = self_keys - self.REQUIRED_KEYS
        if missing_keys:
            return False
        return True


class PdfPage(RestrictedDict):
    REQUIRED_KEYS = frozenset(['/Type', '/Parent', '/Resources', '/MediaBox'])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def merge(self, other):
        if not isinstance(other, PdfPage):
            raise ValueError("PdfPage can only be merged with another PdfPage")


class Resources(PdfDict):
    pass








class Pdf(qpdf.QPDF):

    @property
    def root(self):
        return PdfObject.specialize(super().root)

    @property
    def trailer(self):
        return super().trailer



import pytest
from pikepdf import Pdf
import os
import platform
import shutil
from pathlib import Path
from contextlib import suppress
from subprocess import run, PIPE, STDOUT, DEVNULL
import xml.etree.ElementTree as ET


VERAPDF = Path(os.environ['HOME']) / 'verapdf' / 'verapdf'


NO_PDFA_VALIDATOR = not VERAPDF.is_file()


def verapdf_validate(filename):
    with open(filename, 'rb') as f:
        proc = run([VERAPDF], stdin=f, stdout=PIPE, stderr=STDOUT, check=True)
        result = proc.stdout.decode('utf-8')

        xml_start = result.find('<?xml version')
        xml = result[xml_start:]

    root = ET.fromstring(xml)
    node = root.find(".//taskResult[@type='VALIDATE']")
    return node.attrib['isExecuted'] == 'true' and \
            node.attrib['isSuccess'] == 'true'


@pytest.mark.skipif(NO_PDFA_VALIDATOR, reason="can't find verapdf")
def test_pdfa_sanity(resources, outdir):
    filename = resources / 'veraPDF test suite 6-2-10-t02-pass-a.pdf'

    assert verapdf_validate(filename)

    pdf = Pdf.open(filename)
    pdf.save(outdir / 'pdfa.pdf')

    assert verapdf_validate(outdir / 'pdfa.pdf')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)

from io import BytesIO
from subprocess import run, PIPE
from tempfile import NamedTemporaryFile
from itertools import zip_longest
import struct

from ._objects import Name
from . import Pdf, Object, ObjectType, Array

class DependencyError(Exception):
    pass

class UnsupportedImageTypeError(Exception):
    pass


class _PdfImageDescriptor:
    def __init__(self, name, type_, default, inline_name=None, inline_map=None):
        self.name = name
        self.type = type_
        self.default = default
        self.inline_name = inline_name
        self.inline_map = inline_map

    def __get__(self, wrapper, wrapperclass):
        sentinel = object()
        val = sentinel
        if self.inline_name:
            val = getattr(wrapper.obj, self.inline_name, sentinel)
        if val is sentinel:
            val = getattr(wrapper.obj, self.name, self.default)
        if self.type == bool:
            return val.as_bool() if isinstance(val, Object) else bool(val)
        return self.type(val)

    def __set__(self, wrapper, val):
        if self.inline_name:
            raise NotImplementedError("editing inline images")
        setattr(wrapper.obj, self.name, val)


def array_str(value):
    if isinstance(value, list):
        return [str(item) for item in value]
    if value.type_code == ObjectType.array:
        return [str(item) for item in value]
    elif value.type_code == ObjectType.name:
        return [str(value)]
    raise NotImplementedError(value)


def dict_or_array_dict(value):
    if isinstance(value, list):
        return value
    if value.type_code == ObjectType.dictionary:
        return [value.as_dict()]
    elif value.type_code == ObjectType.array:
        return [v.as_dict() for v in value]


class PdfImage:
    SIMPLE_COLORSPACES = ('/DeviceRGB', '/DeviceGray', '/CalRGB', '/CalGray')

    def __init__(self, obj):
        if obj.type_code == ObjectType.stream and \
                obj.stream_dict.get("/Subtype") != "/Image":
            raise TypeError("can't construct PdfImage from non-image")
        self.obj = obj

    width = _PdfImageDescriptor('Width', int, None)
    height = _PdfImageDescriptor('Height', int, None)
    image_mask = _PdfImageDescriptor('ImageMask', bool, False)
    _bpc = _PdfImageDescriptor('BitsPerComponent', int, None)
    _colorspaces = _PdfImageDescriptor('ColorSpace', array_str, [])
    filters = _PdfImageDescriptor('Filter', array_str, [])
    decode_parms = _PdfImageDescriptor('DecodeParms', dict_or_array_dict, [])

    @property
    def bits_per_component(self):
        if self._bpc is None:
            return 1 if self.image_mask else 8
        return self._bpc

    @bits_per_component.setter
    def bits_per_component(self, val):
        self._bpc = val

    @property
    def colorspace(self):
        if self.image_mask:
            return None  # Undefined for image masks
        if self._colorspaces[0] in self.SIMPLE_COLORSPACES:
            return self._colorspaces[0]
        if self._colorspaces[0] == '/DeviceCMYK':
            return self._colorspaces[0]
        if self._colorspaces[0] == '/Indexed' \
                and self._colorspaces[1] in self.SIMPLE_COLORSPACES:
            return self._colorspaces[1]
        if self._colorspaces[0] == '/ICCBased':
            icc = self.obj.ColorSpace[1]
            return icc.stream_dict.get('/Alternate', '')
        raise NotImplementedError("not sure how to get colorspace")

    @property
    def is_inline(self):
        return False

    @property
    def indexed(self):
        return self._colorspaces[0] == '/Indexed'

    @property
    def palette(self):
        if not self.indexed:
            return None
        idx, base, hival, lookup = None, None, None, None
        try:
            idx, base, hival, lookup = self.obj.ColorSpace.as_list()
        except ValueError as e:
            raise ValueError('Not sure how to interpret this palette') from e
        base = str(base)
        hival = int(hival)
        lookup = bytes(lookup)
        if not base in self.SIMPLE_COLORSPACES:
            raise NotImplementedError("not sure how to interpret this palette")
        if base == '/DeviceRGB':
            base = 'RGB'
        elif base == '/DeviceGray':
            base = 'L'
        return base, lookup

    @property
    def size(self):
        return self.width, self.height

    @property
    def mode(self):
        m = ''
        if self.indexed:
            m = 'P'
        elif self.bits_per_component == 1:
            m = '1'
        elif self.bits_per_component == 8:
            if self.colorspace == '/DeviceRGB':
                m = 'RGB'
            elif self.colorspace == '/DeviceGray':
                m = 'L'
        if m == '':
            raise NotImplementedError("Not sure how to handle PDF image of this type")
        return m

    @property
    def filter_decodeparms(self):
        """
        PDF has a lot of optional data structures concerning /Filter and
        /DecodeParms. /Filter can be absent or a name or an array, /DecodeParms
        can be absent or a dictionary (if /Filter is a name) or an array (if
        /Filter is an array). When both are arrays the lengths match.

        Normalize this into:
        [(/FilterName, {/DecodeParmName: Value, ...}), ...]

        The order of /Filter matters as indicates the encoding/decoding sequence.

        """
        return list(zip_longest(self.filters, self.decode_parms, fillvalue={}))

    def extract_to(self, *, stream):
        """
        Attempt to extract the image directly to a usable image file

        If there is no way to extract the image without decompressing or
        transcoding then raise an exception. The type and format of image
        generated will vary.

        :param stream: Writable stream to write data to
        """

        if self.filters == ['/CCITTFaxDecode']:
            data = self.obj.read_raw_bytes()
            stream.write(self._generate_ccitt_header(data))
            stream.write(data)
            return '.tif'
        elif self.filters == ['/DCTDecode'] and \
                self.mode == 'RGB' and \
                self.filter_decodeparms[0][1].get('/ColorTransform', 1):
            buffer = self.obj.get_raw_stream_buffer()
            stream.write(buffer)
            return '.jpg'

        raise UnsupportedImageTypeError()

    def read_bytes(self):
        return self.obj.read_bytes()

    def get_stream_buffer(self):
        return self.obj.get_stream_buffer()

    def as_pil_image(self):
        """
        Extract the image as a Pillow Image, using decompression as necessary

        """
        from PIL import Image

        try:
            bio = BytesIO()
            self.extract_to(stream=bio)
            bio.seek(0)
            return Image.open(bio)
        except UnsupportedImageTypeError:
            pass

        if self.mode == 'RGB':
            # No point in accessing the buffer here, size qpdf decodes to 3-byte
            # RGB and Pillow needs RGBX for raw access
            data = self.read_bytes()
            im = Image.frombytes('RGB', self.size, data)
        elif self.mode in ('L', 'P'):
            buffer = self.get_stream_buffer()
            stride = 0  # tell Pillow to calculate stride from line width
            ystep = 1  # image is top to bottom in memory
            im = Image.frombuffer('L', self.size, buffer, "raw", 'L', stride,
                                  ystep)
            if self.mode == 'P':
                base_mode, palette_data = self.palette
                if base_mode in ('RGB', 'L'):
                    im.putpalette(palette_data, rawmode=base_mode)
                else:
                    raise NotImplementedError('palette with ' + base_colorspace)

        if not im:
            raise UnsupportedImageTypeError(repr(self))

        return im

    def _generate_ccitt_header(self, data):
        # https://stackoverflow.com/questions/2641770/
        # https://www.itu.int/itudoc/itu-t/com16/tiff-fx/docs/tiff6.pdf

        if not self.decode_parms:
            raise ValueError("/CCITTFaxDecode without /DecodeParms")

        if self.decode_parms[0].get("/K", 1) < 0:
            ccitt_group = 4  # Pure two-dimensional encoding (Group 4)
        else:
            ccitt_group = 3
        black_is_one = self.decode_parms[0].get("/BlackIs1", False)
        white_is_zero = 1 if black_is_one else 0

        img_size = len(data)
        tiff_header_struct = '<' + '2s' + 'H' + 'L' + 'H' + 'HHLL' * 8 + 'L'
        tiff_header = struct.pack(
            tiff_header_struct,
            b'II',  # Byte order indication: Little endian
            42,  # Version number (always 42)
            8,  # Offset to first IFD
            8,  # Number of tags in IFD
            256, 4, 1, self.width,  # ImageWidth, LONG, 1, width
            257, 4, 1, self.height,  # ImageLength, LONG, 1, length
            258, 3, 1, 1,  # BitsPerSample, SHORT, 1, 1
            259, 3, 1, ccitt_group,  # Compression, SHORT, 1, 4 = CCITT Group 4 fax encoding
            262, 3, 1, int(white_is_zero),  # Thresholding, SHORT, 1, 0 = WhiteIsZero
            273, 4, 1, struct.calcsize(tiff_header_struct),  # StripOffsets, LONG, 1, length of header
            278, 4, 1, self.height,
            279, 4, 1, img_size,  # StripByteCounts, LONG, 1, size of image
            0  # last IFD
        )
        return tiff_header

    def show(self):
        self.as_pil_image().show()

    def __repr__(self):
        return '<pikepdf.PdfImage image mode={} size={}x{} at {}>'.format(
            self.mode, self.width, self.height, hex(id(self)))

    def _repr_png_(self):
        "Display hook for IPython/Jupyter"
        b = BytesIO()
        im = self.as_pil_image()
        im.save(b, 'PNG')
        return b.getvalue()


def inline_remove_abbrevs(value):
    abbrevs = {
        '/G': '/DeviceGray',
        '/RGB': '/DeviceRGB',
        '/CMYK': '/DeviceCMYK',
        '/I': '/Indexed',
        '/AHx': '/ASCIIHexDecode',
        '/A85': '/ASCII85Decode',
        '/LZW': '/LZWDecode',
        '/RL': '/RunLengthDecode',
        '/CCF': '/CCITTFaxDecode',
        '/DCT': '/DCTDecode'
    }
    return [abbrevs.get(value, value) for value in array_str(value)]


class PdfInlineImage(PdfImage):

    def __init__(self, *, image_data, image_object: tuple):
        self._data = image_data
        self._image_object = image_object

        reparse = ' '.join([obj.unparse_resolved() for obj in image_object])
        self.obj = Object.parse(('<< ' + reparse + ' >>').encode('ascii'))

    width = _PdfImageDescriptor('Width', int, None, 'W')
    height = _PdfImageDescriptor('Height', int, None, 'H')
    image_mask = _PdfImageDescriptor('ImageMask', bool, False, 'IM')
    _bpc = _PdfImageDescriptor('BitsPerComponent', int, None, 'BPC')
    _colorspaces = _PdfImageDescriptor('ColorSpace', inline_remove_abbrevs, [], 'CS')
    filters = _PdfImageDescriptor('Filter', inline_remove_abbrevs, [], 'F')
    decode_parms = _PdfImageDescriptor('DecodeParms', dict_or_array_dict, [], 'DP')

    @property
    def is_inline(self):
        return True

    def __repr__(self):
        return '<pikepdf.PdfInlineImage image mode={} size={}x{} at {}>'.format(
            self.mode, self.width, self.height, hex(id(self)))

    def extract_to(self, *, stream):
        raise UnsupportedImageTypeError("inline images don't support extract")

    def read_bytes(self):
        raise NotImplementedError("qpdf returns compressed")
        #return self._data._inline_image_bytes()

    def get_stream_buffer(self):
        raise NotImplementedError("qpdf returns compressed")
        #return memoryview(self._data.inline_image_bytes())


def page_to_svg(page):
    pdf = Pdf.new()
    pdf.pages.append(page)
    with NamedTemporaryFile(suffix='.pdf') as tmp_in, \
            NamedTemporaryFile(mode='w+b', suffix='.svg') as tmp_out:
        pdf.save(tmp_in)
        tmp_in.seek(0)

        try:
            proc = run(['mudraw', '-F', 'svg', '-o', tmp_out.name, tmp_in.name], stderr=PIPE)
        except FileNotFoundError as e:
            raise DependencyError("Could not find the required executable 'mutool'")

        if proc.stderr:
            print(proc.stderr.decode())
        tmp_out.flush()
        tmp_out2 = open(tmp_out.name, 'rb')  # Not sure why re-opening is need, but it is
        svg = tmp_out2.read()
        return svg.decode()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)

"""Support functions called by the C++ library binding layer

"""

import re
from itertools import tee
import os
import sys
from collections.abc import MutableSequence


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)

# Provide os.fspath equivalent for Python <3.6
if sys.version_info[0:2] <= (3, 5):
    def fspath(path):
        import pathlib
        '''https://www.python.org/dev/peps/pep-0519/#os'''
        if isinstance(path, (str, bytes)):
            return path

        # Work from the object's type to match method resolution of other magic
        # methods.
        path_type = type(path)
        try:
            path = path_type.__fspath__(path)
        except AttributeError:
            # Added for Python 3.5 support.
            if isinstance(path, pathlib.Path):
                return str(path)
            elif hasattr(path_type, '__fspath__'):
                raise
        else:
            if isinstance(path, (str, bytes)):
                return path
            else:
                raise TypeError("expected __fspath__() to return str or bytes, "
                                "not " + type(path).__name__)

        raise TypeError(
            "expected str, bytes, pathlib.Path or os.PathLike object, not "
            + path_type.__name__)

else:
    fspath = os.fspath



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)

"""Provide classes to stand in for PDF objects

The purpose of these is to provide nice-looking classes to allow explicit
construction of PDF objects and more pythonic idioms and facilitate discovery
by documentation generators.

It's also a place to narrow the scope of input types to those more easily
converted to C++.

In reality all of these return objects of class pikepdf.Object or rather
QPDFObjectHandle which is a generic type.

"""

from decimal import Decimal, InvalidOperation
from math import isfinite

from . import _qpdf
from ._qpdf import Object, ObjectType, Stream, Operator, Null


class _ObjectMeta(type):
    "Supports instance checking"

    def __instancecheck__(cls, instance):
        if type(instance) != Object:
            return False
        return cls.object_type == instance.type_code


class Boolean(metaclass=_ObjectMeta):
    object_type = ObjectType.boolean

    def __new__(cls, value):
        return _qpdf._new_boolean(value)


class Integer(metaclass=_ObjectMeta):
    object_type = ObjectType.integer

    def __new__(cls, n):
        if n.bit_length() >= 64:
            raise ValueError('Value is too large for 64-bit integer')
        if not isinstance(n, int) and n == int(n):
            n = int(n)
        return _qpdf._new_integer(n)


class Real(metaclass=_ObjectMeta):
    object_type = ObjectType.real

    def __new__(cls, value, dec_places=0):
        if dec_places < 0 or not isinstance(dec_places, int):
            raise ValueError('dec_places must be nonnegative integer')

        if isinstance(value, int):
            return _qpdf._new_real(value, 0)

        if isinstance(value, float) and isfinite(value):
            return _qpdf._new_real(value, dec_places)

        try:
            dec = Decimal(value)
        except InvalidOperation:
            raise TypeError('Could not convert object to int, float or Decimal')

        if dec.is_infinite() or dec.is_nan():
            raise ValueError('NaN and infinity are not valid PDF objects')

        return _qpdf._new_real(str(dec))


class Name(metaclass=_ObjectMeta):
    object_type = ObjectType.name

    def __new__(cls, name):
        # QPDF_Name::unparse ensures that names are always saved in a UTF-8
        # compatible way, so we only need to guard the input.
        if isinstance(name, bytes):
            raise TypeError("Name should be str")
        return _qpdf._new_name(name)


class String(metaclass=_ObjectMeta):
    object_type = ObjectType.string

    def __new__(cls, s):
        if isinstance(s, bytes):
            return _qpdf._new_string(s)
        try:
            ascii = s.encode('ascii')
            return _qpdf._new_string(ascii)
        except UnicodeEncodeError:
            utf16 = b'\xfe\xff' + s.encode('utf-16be')
            return _qpdf._new_string(utf16)


class Array(metaclass=_ObjectMeta):
    object_type = ObjectType.array

    def __new__(cls, a):
        if isinstance(a, (str, bytes)):
            raise TypeError('Strings cannot be converted to arrays of chars')
        return _qpdf._new_array(a)


class Dictionary(metaclass=_ObjectMeta):
    object_type = ObjectType.dictionary

    def __new__(cls, d):
        return _qpdf._new_dictionary(d)
